"""Python Package for controlling Alexa devices (echo dot, etc) programmatically.

SPDX-License-Identifier: Apache-2.0

This provides a login by proxy method.
Built on https://github.com/alandtse/auth_capture_proxy

For more details about this api, please refer to the documentation at
https://gitlab.com/keatontaylor/alexapy
"""

import logging
from functools import partial

import authcaptureproxy
from bs4 import BeautifulSoup
from yarl import URL

from alexapy.alexalogin import AlexaLogin
from alexapy.helpers import hide_email, hide_password

_LOGGER = logging.getLogger(__name__)


class AlexaProxy(authcaptureproxy.AuthCaptureProxy):
    """Class to handle proxy login connections to Alexa."""

    def __init__(self, login: AlexaLogin, base_url: str) -> None:
        """Initialize proxy object.

        Args:
            login (AlexaLogin): AlexaLogin object to update after completion of proxy.
            base_url (Text): base url for proxy location. e.g., http://192.168.1.1

        """
        super().__init__(URL(base_url), URL(login.start_url))
        self._login: AlexaLogin = login
        self._config_flow_id = None
        self._callback_url = None
        self.tests = {"test_amazon_url": self.test_amazon_url}
        self.modifiers = {
            "autofill": partial(
                self.autofill,
                {
                    "email": self._login.email,
                    "password": self._login.password,
                    "otpCode": self._login.get_totp_token(),
                },
            )
        }

    async def test_amazon_url(self, resp, data, query) -> URL | str | None:
        """Test for Alexa success.

        Args
            resp (httpx.Response): The aiohttp response.
            data (Dict[Text, Any]): Dictionary of all post data captured through proxy with overwrites for duplicate keys.
            query (Dict[Text, Any]): Dictionary of all query data with overwrites for duplicate keys.

        Returns
            Optional[Union[URL, Text]]: URL for a http 302 redirect or Text to display on success. None indicates test did not pass.

        """  # noqa: E501
        if URL(str(resp.url)).path in ["/ap/maplanding", "/spa/index.html"]:
            self._login.session.cookie_jar.update_cookies(self.session.cookies)
            self._login.authorization_code = URL(str(resp.url)).query.get(
                "openid.oa2.authorization_code"
            )
            self._login.access_token = URL(str(resp.url)).query.get(
                "openid.oa2.access_token"
            )
            self._config_flow_id = self.init_query.get("config_flow_id")
            self._callback_url = self.init_query.get("callback_url")
            # Reset so proxy can be reused
            # Unfortunately because the route has a handler bound and cannot be changed, we need to remove the modifiers for other proxies. # noqa: E501
            await self.reset_data()
            if self._callback_url:
                return URL(self._callback_url)
            return f"Successfully logged in as {self._login.email} for flow {self._config_flow_id}. Please close the window."  # noqa: E501
        if URL(str(resp.url)).path in [
            "/"
        ]:  # sometimes the redirect after the captcha fails, redirect manually
            return self._login.proxy_url

    def change_login(self, login: AlexaLogin) -> None:
        """Change login.

        Args:
            login (AlexaLogin): AlexaLogin object to update after completion of proxy.
        """
        self._login = login
        self.tests = {"test_amazon_url": self.test_amazon_url}
        self.modifiers = {
            "autofill": partial(
                self.autofill,
                {
                    "email": self._login.email,
                    "password": self._login.password,
                    "otpCode": self._login.get_totp_token(),
                },
            )
        }

    def autofill(self, items: dict, html: str) -> str:
        """Autofill input tags in form in html.

        Args
            html (Text): html to convert
            items (dict): Dictionary of values to fill

        Returns
            Text: html with values filled in

        """
        soup: BeautifulSoup = BeautifulSoup(html, "html.parser")
        for item, value in items.items():
            for html_tag in soup.find_all(attrs={"name": item}):
                if not html_tag.get("value"):
                    html_tag["value"] = value
                    if self._login._debug:
                        _LOGGER.debug(
                            "Filled %s",
                            str(html_tag).replace(
                                value,
                                (
                                    hide_password(value)
                                    if item == "password"
                                    else hide_email(value)
                                    if item == "email"
                                    else value
                                ),
                            ),
                        )
        return str(soup)



"""Python Package for controlling Alexa devices (echo dot, etc) programmatically.

SPDX-License-Identifier: Apache-2.0

Websocket library.

This library is based on MIT code from https://github.com/Apollon77/alexa-remote.

For more details about this api, please refer to the documentation at
https://gitlab.com/keatontaylor/alexapy
"""

import asyncio
import contextlib
import datetime
import json
import logging
import ssl
from collections.abc import Callable, Coroutine
from typing import Any

import certifi
import httpx

from alexapy.const import HTTP2_AUTHORITY, HTTP2_DEFAULT
from alexapy.errors import AlexapyLoginError

from .alexalogin import AlexaLogin


def _create_ssl_context():
    """Create SSL context."""
    context = ssl.create_default_context(
        purpose=ssl.Purpose.SERVER_AUTH, cafile=certifi.where()
    )
    return context


_SSL_CONTEXT = _create_ssl_context()
_LOGGER = logging.getLogger(__name__)


class HTTP2EchoClient:
    """HTTP2 Client Class for Echo Devices.

    Based on code from openHAB:
    https://github.com/Apollon77/alexa-remote/blob/bc687b9e36da7c2318c56b4e1bec677c7198dbd4/alexa-http2push.js
    """

    def __init__(
        self,
        login: AlexaLogin,
        msg_callback: Callable[[Any], Coroutine[Any, Any, None]],
        open_callback: Callable[[], Coroutine[Any, Any, None]],
        close_callback: Callable[[], Coroutine[Any, Any, None]],
        error_callback: Callable[[str], Coroutine[Any, Any, None]],
        loop: asyncio.AbstractEventLoop | None = None,
    ) -> None:
        """Init for threading and HTTP2 Push Connection."""
        assert login.session is not None
        self._options = {
            "method": "GET",
            "path": "/v20160207/directives",
            "authority": HTTP2_AUTHORITY.get(
                login.url.replace("amazon", ""), HTTP2_DEFAULT
            ),
            "scheme": "https",
            "authorization": f"Bearer {login.access_token}",
        }
        self.open_callback: Callable[[], Coroutine[Any, Any, None]] = open_callback
        self.msg_callback: Callable[[Any], Coroutine[Any, Any, None]] = msg_callback
        self.close_callback: Callable[[], Coroutine[Any, Any, None]] = close_callback
        self.error_callback: Callable[[str], Coroutine[Any, Any, None]] = error_callback
        self._http2url: str = (
            f"https://{self._options['authority']}{self._options['path']}"
        )
        self.client = httpx.AsyncClient(http2=True, verify=_SSL_CONTEXT)
        self.boundary: str = ""
        self._login = login
        self._loop: asyncio.AbstractEventLoop = (
            loop if loop else asyncio.get_event_loop()
        )
        self._last_ping = datetime.datetime(1, 1, 1)
        self._tasks = set()

    async def async_run(self) -> None:
        """Start Async WebSocket Listener."""
        task = self._loop.create_task(self.process_messages())
        task.add_done_callback(self.on_close)
        self._tasks.add(task)
        task = self._loop.create_task(self.manage_pings())
        task.add_done_callback(self.on_close)
        self._tasks.add(task)
        # task = self._loop.create_task(self.test_close(raise_exception=True))
        # task.add_done_callback(self.on_close)
        # self._tasks.add(task)
        await self.async_on_open()

    async def process_messages(self) -> None:
        """Start Async WebSocket Listener."""
        _LOGGER.debug("Starting message parsing loop.")
        _LOGGER.debug("Connecting to %s with %s", self._http2url, self._options)
        try:
            async with self.client.stream(
                self._options["method"],
                self._http2url,
                headers={
                    "authorization": self._options["authorization"],
                },
                timeout=httpx.Timeout(None),
            ) as response:
                async for data in response.aiter_text():
                    await self.on_message(data)
        except httpx.RemoteProtocolError as exception_:
            self.on_close(f"Disconnect detected: {exception_}")

    async def on_message(self, message: str) -> None:
        """Handle New Message."""
        reauth_required = "Unable to authenticate the request. Please provide a valid authorization token."  # noqa: E501
        _LOGGER.debug("Received raw message: %s", message)
        for line in message.splitlines():
            if line.startswith("------"):
                if not self.boundary:  # set boundary character
                    self.boundary = line
            elif line.startswith(reauth_required):
                _LOGGER.debug("HTTP2 login error: %s", message)
                await self.handle_login_error("HTTP2 Message Parsing reauth")
            elif line.startswith("Content-Type:"):
                continue
            elif line and not line.startswith(self.boundary):
                with contextlib.suppress(json.decoder.JSONDecodeError):
                    asyncio.run_coroutine_threadsafe(
                        self.msg_callback(json.loads(line)), self._loop
                    )

    def on_error(self, error: str = "Unspecified") -> None:
        """Handle HTTP2 Error."""
        _LOGGER.debug("HTTP2 Error: %s", error)
        asyncio.run_coroutine_threadsafe(self.error_callback(error), self._loop)

    def on_close(self, future="") -> None:
        """Handle HTTP2 Close."""
        _LOGGER.debug("HTTP2 Connection Closed.")
        asyncio.gather(*self._tasks, return_exceptions=True)
        try:
            for task in self._tasks:
                if task.done():
                    exc = task.exception()
                    if exc:
                        self.on_error(exc)
                else:
                    task.remove_done_callback(self.on_close)
                    task.cancel()
        except BaseException:
            pass
        asyncio.run_coroutine_threadsafe(self.close_callback(), self._loop)

    async def async_on_open(self) -> None:
        """Handle Async WebSocket Open."""
        await self.open_callback()

    async def manage_pings(self) -> None:
        """Manage Pings."""
        while True:
            await self.ping()
            await asyncio.sleep(299)

    async def ping(self) -> None:
        """Ping."""
        url = f"https://{self._options['authority']}/ping"
        _LOGGER.debug("Preparing ping to %s", url)
        response = await self.client.get(
            url,
            headers={
                "authorization": self._options["authorization"],
            },
        )
        self._last_ping = datetime.datetime.now()
        _LOGGER.debug(
            "Received response: %s:%s",
            response.status_code,
            response.text,
        )
        if response.status_code in [403]:
            _LOGGER.debug("Detected ping 403")
            await self.handle_login_error(response.text)

    async def handle_login_error(self, message: str = "") -> None:
        """Handle login error.

        Attempt to relogin otherwise raise login error.
        """
        if not await self._login.test_loggedin():
            raise AlexapyLoginError(message)

    async def test_close(self, delay: int = 30, raise_exception: bool = False) -> None:
        """Test close."""
        await asyncio.sleep(delay)
        if raise_exception:
            raise AlexapyLoginError(f"Raised exception after {delay}")



"""Python Package for controlling Alexa devices (echo dot, etc) programmatically.

SPDX-License-Identifier: Apache-2.0

Helpers.

For more details about this api, please refer to the documentation at
https://gitlab.com/keatontaylor/alexapy
"""

import functools
import logging
import os
from asyncio import CancelledError
from http.cookies import CookieError
from json import JSONDecodeError
from types import MappingProxyType
from typing import Any

import aiofiles.os as aioos
from aiohttp import (
    ClientConnectionError,
    ClientResponse,
    ContentTypeError,
    ServerDisconnectedError,
)

import alexapy.alexalogin

from .const import EXCEPTION_TEMPLATE
from .errors import (
    AlexapyConnectionError,
    AlexapyLoginCloseRequested,
    AlexapyLoginError,
)

_LOGGER = logging.getLogger(__name__)

MIN_HIDE_LENGTH = 6

# Alexa API: Successful but guaranteed to NOT contain JSON
NO_JSON_STATUSES = (204,)


def hide_email(email: str) -> str:
    """Obfuscate email."""
    part = email.split("@")
    if len(part) > 1:
        return f"{part[0][0]}{'*' * (len(part[0]) - 2)}{part[0][-1]}@{part[1][0]}{'*' * (len(part[1]) - 2)}{part[1][-1]}"  # noqa: E501
    return hide_serial(email)


def hide_password(value: str) -> str:
    """Obfuscate password."""
    return f"REDACTED {len(value)} CHARS"


def hide_serial(item: dict | str | list | None) -> dict | str | list:
    """Obfuscate serial."""
    if item is None:
        return ""
    if isinstance(item, dict):
        response = item.copy()
        for key, value in item.items():
            if (
                isinstance(value, (dict, list))
                or key
                in [
                    "deviceSerialNumber",
                    "serialNumber",
                    "destinationUserId",
                    "customerId",
                    "access_token",
                    "refresh_token",
                ]
                or "secret" in key
            ):
                response[key] = hide_serial(value)
    elif isinstance(item, str):
        response = (
            f"{item[0]}{'*' * (len(item) - 4)}{item[-3:]}"
            if len(item) > MIN_HIDE_LENGTH
            else f"{'*' * len(item)}"
        )
    elif isinstance(item, list):
        response = []
        for list_item in item:
            if isinstance(list_item, dict):
                response.append(hide_serial(list_item))
            else:
                response.append(list_item)
    return response


def obfuscate(item):
    """Obfuscate email, password, and other known sensitive keys."""
    if item is None:
        return ""
    if isinstance(item, (MappingProxyType, dict)):
        response = item.copy()
        for key, value in item.items():
            if key in ["password"]:
                response[key] = hide_password(value)
            elif key in ["email"]:
                response[key] = hide_email(value)
            elif key in ["cookies_txt"]:
                response[key] = "OBFUSCATED COOKIE"
            elif (
                key
                in [
                    "deviceSerialNumber",
                    "serialNumber",
                    "destinationUserId",
                    "customerId",
                    "access_token",
                    "refresh_token",
                ]
                or "secret" in key
            ):
                response[key] = hide_serial(value)
            elif isinstance(value, (dict, list, tuple)):
                response[key] = obfuscate(value)
    elif isinstance(item, (list, tuple)):
        response = []
        for list_item in item:
            if isinstance(list_item, (dict, list, tuple)):
                response.append(obfuscate(list_item))
            else:
                response.append(list_item)
        if isinstance(item, tuple):
            response = tuple(response)
    else:
        return item
    return response


def _catch_all_exceptions(func):
    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        login: alexapy.alexalogin.AlexaLogin | None = None
        for arg in args:
            if isinstance(arg, alexapy.alexalogin.AlexaLogin):
                login = arg
                break
        try:
            return await func(*args, **kwargs)
        except (ClientConnectionError, KeyError, ServerDisconnectedError) as ex:
            _LOGGER.warning(
                "%s.%s(%s, %s): A connection error occurred: %s",
                func.__module__[func.__module__.find(".") + 1 :],
                func.__name__,
                obfuscate(args),
                obfuscate(kwargs),
                EXCEPTION_TEMPLATE.format(type(ex).__name__, ex.args),
            )
            raise AlexapyConnectionError from ex
        except (JSONDecodeError, CookieError) as ex:
            _LOGGER.warning(
                "%s.%s(%s, %s): A login error occurred: %s",
                func.__module__[func.__module__.find(".") + 1 :],
                func.__name__,
                obfuscate(args),
                obfuscate(kwargs),
                EXCEPTION_TEMPLATE.format(type(ex).__name__, ex.args),
            )
            if login:
                login.status["login_successful"] = False
            raise AlexapyLoginError from ex
        except ContentTypeError as ex:
            _LOGGER.warning(
                "%s.%s(%s, %s): A login error occurred; Amazon may want you to change your password: %s",  # noqa: E501
                func.__module__[func.__module__.find(".") + 1 :],
                func.__name__,
                obfuscate(args),
                obfuscate(kwargs),
                EXCEPTION_TEMPLATE.format(type(ex).__name__, ex.args),
            )
            if login:
                login.status["login_successful"] = False
            raise AlexapyLoginError from ex
        except CancelledError as ex:
            _LOGGER.warning(
                "%s.%s(%s, %s): Timeout error occurred accessing AlexaAPI: %s",
                func.__module__[func.__module__.find(".") + 1 :],
                func.__name__,
                obfuscate(args),
                obfuscate(kwargs),
                EXCEPTION_TEMPLATE.format(type(ex).__name__, ex.args),
            )
            return None
        except AlexapyLoginCloseRequested:
            raise
        except Exception as ex:
            _LOGGER.warning(
                "%s.%s(%s, %s): An error occurred accessing AlexaAPI: %s",
                func.__module__[func.__module__.find(".") + 1 :],
                func.__name__,
                obfuscate(args),
                obfuscate(kwargs),
                EXCEPTION_TEMPLATE.format(type(ex).__name__, ex.args),
            )
            raise
            # return None

    return wrapper


async def delete_cookie(cookiefile: str) -> None:
    """Delete a cookie.

    Args:
        cookiefile (Text): Path to cookie

    """
    _LOGGER.debug("Deleting cookiefile %s ", cookiefile)
    try:
        try:
            await aioos.remove(cookiefile)
        except AttributeError:
            os.remove(cookiefile)
    except (OSError, EOFError, TypeError, AttributeError) as ex:
        _LOGGER.debug(
            "Error deleting cookie: %s; please manually remove",
            EXCEPTION_TEMPLATE.format(type(ex).__name__, ex.args),
        )


async def get_json_value(
    response: ClientResponse | None,
    path: str | None,
    expected_type: type | tuple[type, ...],
) -> tuple[Any, bool]:
    """Extract a value from a JSON response using a dot-separated path.

    Retrieves the JSON body from an aiohttp.ClientResponse and traverses
    it according to a dot-path (e.g., 'items.0.name'). Supports array
    indices, multiple expected types, and handles Alexa API response behaviors.

    If `path` is None or an empty string, the entire JSON body is returned.

    Args:
        response: The aiohttp.ClientResponse object, or None.
        path: Dot-separated path to the desired value in the JSON. If None
            or empty, returns the entire JSON body.
        expected_type: A type or tuple of types that the value is expected to have.

    Returns:
        A tuple (value, valid):
        - value: the extracted value, or None if missing/invalid
        - valid: True if the value exists and matches the expected type, False otherwise

    Logs warnings for:
        - HTTP statuses that do not contain JSON (e.g., 204)
        - HTTP error statuses (4xx/5xx)
        - Missing or out-of-range path elements
        - Type mismatches

    """
    if response is None:
        return None, False

    body = None
    value = None
    valid = True  # Tracks overall success
    status = response.status

    # Status that definitely contains no JSON (e.g., 204)
    if status in NO_JSON_STATUSES:
        _LOGGER.warning(
            "HTTP (URL: %s) status %s returned but contains no JSON. Cannot validate '%s'.",  # noqa: E501
            response.url,
            status,
            path,
        )
        valid = False

    # HTTP error status (4xx/5xx)
    elif status >= 400:
        _LOGGER.warning(
            "HTTP (URL: %s) error status %s when accessing JSON for '%s'.",
            response.url,
            status,
            path,
        )
        valid = False

    # Status that may contain JSON (200/201/202)
    else:
        try:
            body = await response.json(content_type=None)
            value = body
        except Exception as exc:
            _LOGGER.warning(
                "Failed (URL: %s) to parse JSON for '%s' (status %s): %s",
                response.url,
                path,
                status,
                exc,
            )
            _LOGGER.debug("Response text: %s", response.text)
            valid = False

        # Traverse dot-path only if still valid
        if valid and path is not None and path != "":
            for part in path.split("."):
                # List index access
                if isinstance(value, list) and part.isdigit():
                    idx = int(part)
                    if idx < 0 or idx >= len(value):
                        _LOGGER.warning(
                            "Resulting JSON (URL: %s), index %s out of range at '%s'.",
                            response.url,
                            part,
                            path,
                        )
                        _LOGGER.debug("Response json: %s", body)
                        valid = False
                        value = None
                        break
                    value = value[idx]
                    continue

                # Dict key access
                if isinstance(value, dict) and part in value:
                    value = value[part]
                    continue

                # Path not found
                _LOGGER.warning(
                    "Resulting JSON (URL: %s), path '%s' not found in JSON response.",
                    response.url,
                    path,
                )
                _LOGGER.debug("Response json: %s", body)
                valid = False
                value = None
                break

        # Type check
        if valid and not isinstance(value, expected_type):
            type_names = (
                ", ".join(t.__name__ for t in expected_type)
                if isinstance(expected_type, tuple)
                else expected_type.__name__
            )
            _LOGGER.warning(
                "Resulting JSON (URL: %s), invalid type at '%s': expected %s, got %s",
                response.url,
                path,
                type_names,
                type(value).__name__,
            )
            _LOGGER.debug("Response json: %s", body)
            valid = False
            value = None

    return value, valid



"""Python Package for controlling Alexa devices (echo dot, etc) programmatically.

SPDX-License-Identifier: Apache-2.0

API access.

For more details about this api, please refer to the documentation at
https://gitlab.com/keatontaylor/alexapy
"""

import asyncio
import contextlib
import json
import logging
import math
import random
import time
import urllib.parse
from json.decoder import JSONDecodeError
from types import NoneType
from typing import Any, ClassVar

import backoff
from aiohttp import ClientConnectionError, ClientResponse
from yarl import URL

from .alexalogin import AlexaLogin
from .const import (
    ALEXA_API_BASE,
    API_USER_AGENT,
    CALL_VERSION,
    DEFAULT_ACCEPT_LANGUAGE,
    DEFAULT_LOCALE,
    GQL_SMARTHOME_QUERY,
)
from .errors import (
    AlexapyConnectionError,
    AlexapyLoginCloseRequested,
    AlexapyLoginError,
    AlexapyTooManyRequestsError,
)
from .helpers import _catch_all_exceptions, get_json_value, hide_email, hide_serial

_LOGGER = logging.getLogger(__name__)


def _min_expo_wait(min_wait: float):
    """Exponential backoff with a specified minimum wait time."""

    def f(*args, **kwargs):
        gen = backoff.expo(*args, **kwargs)
        while True:
            v = next(gen, None)
            # _LOGGER.debug("_min_expo_wait: next(gen) was %s", v)
            v = min_wait if v is None else max(min_wait, v)
            # _LOGGER.debug("_min_expo_wait: returning %s",v)
            yield v

    return f


class AlexaAPI:
    """Class for accessing a specific Alexa device using rest API.

    Args:
    device (AlexaClient): Instance of an AlexaClient to access
    login (AlexaLogin): Successfully logged in AlexaLogin

    """

    devices: ClassVar[dict[str, Any]] = {}
    wake_words: ClassVar[dict[str, Any]] = {}
    _sequence_queue: ClassVar[dict[Any, list[dict[Any, Any]]]] = {}
    _sequence_lock: ClassVar[dict[Any, asyncio.Lock]] = {}

    def __init__(self, device, login: AlexaLogin):
        """Initialize Alexa device."""
        self._device = device
        self._login = login
        self._session = login.session
        self._url: str = "https://alexa." + login.url
        self._login._headers["Referer"] = f"{self._url}/spa/index.html"
        AlexaAPI._sequence_queue[self._login.email] = []
        AlexaAPI._sequence_lock[self._login.email] = asyncio.Lock()
        try:
            csrf = self._login._get_cookies_from_session()["csrf"]
            self._login._headers["csrf"] = csrf.value
        except KeyError as ex:
            _LOGGER.warning(
                (
                    "AlexaLogin session is missing required token: %s "
                    "This may result in authorization errors, please report if unable to login"  # noqa: E501
                ),
                ex,
            )

    def update_login(self, login: AlexaLogin) -> bool:
        """Update Login if it has changed.

        Args
            login (AlexaLogin): AlexaLogin to check

        Returns
            bool: True if change detected

        """
        if login != self._login or login.session != self._session:
            _LOGGER.debug(
                "%s: New Login %s detected; replacing %s",
                hide_email(login.email),
                login,
                self._login,
            )
            self._login = login
            self._session = login.session
            self._url: str = "https://alexa." + login.url
            self._login._headers["Referer"] = f"{self._url}/spa/index.html"
            try:
                csrf = self._login._get_cookies_from_session()["csrf"]
                self._login._headers["csrf"] = csrf.value
            except KeyError as ex:
                if login.status and login.status.get("login_successful"):
                    _LOGGER.warning(
                        (
                            "AlexaLogin session is missing required token: %s "
                            "This may result in authorization errors, please report"
                        ),
                        ex,
                    )
            return True
        return False

    @staticmethod
    async def _get_alexa_api_base(login: AlexaLogin) -> str:
        """Return the base URL for the Alexa REST API (e.g. https://na-api-alexa.amazon.com)."""
        # Cached?
        cached = getattr(login, "_alexa_api_url", None)
        if cached:
            _LOGGER.debug(
                "%s: Returning cached url: %s",
                hide_email(login.email),
                cached
            )
            return cached

        session = login.session
        if session.closed:
            raise AlexapyLoginError("Session is closed")

        # Try /api/endpoints on the web host, e.g. https://alexa.amazon.ca/api/endpoints
        # login.url is usually 'amazon.ca' or 'amazon.com'
        endpoints_url = URL(f"https://alexa.{login.url}/api/endpoints")

        headers = getattr(login, "_headers", {}) or {}
        # Make sure we have some UA + Accept to look like a real client
        local_headers = headers.copy()
        local_headers.setdefault("User-Agent", API_USER_AGENT)
        local_headers.setdefault("Accept", "application/json")

        try:
            resp = await session.get(
                endpoints_url,
                headers=local_headers,
                ssl=login._ssl
            )
            text = await resp.text()
            _LOGGER.debug(
                "%s: /api/endpoints GET %s returned %s:%s:%s",
                hide_email(login.email),
                endpoints_url,
                resp.status,
                resp.reason,
                resp.content_type,
            )
            _LOGGER.debug("%s: Response text: %s", hide_email(login.email), text)
            if resp.status == 200:
                try:
                    data = json.loads(text)
                    api_url = (
                         data.get("websiteApiUrl", "")
                        .rstrip("/")
                    )
                    if api_url:
                        login._alexa_api_url = api_url
                        _LOGGER.debug(
                            "%s: Using websiteApiUrl from /api/endpoints: %s",
                            hide_email(login.email),
                            api_url,
                        )
                        return api_url
                    else:
                        _LOGGER.debug(
                            "%s: Unable to extract websiteApiUrl from /api/endpoints",
                            hide_email(login.email)
                        )
                except ValueError:
                    _LOGGER.debug(
                        "%s: /api/endpoints invalid JSON: %s",
                        hide_email(login.email),
                        hide_serial(text),
                    )
        except ClientConnectionError as exc:
            _LOGGER.debug(
                "%s: /api/endpoints connection error: %s",
                hide_email(login.email),
                exc,
            )

        # Fallback: use region-based default, then web host as last resort
        domain = getattr(login, "url", "")  # e.g. 'amazon.ca'
        suffix = ""
        for sfx in ALEXA_API_BASE:
            if domain.endswith(sfx[1:]):  # strip leading dot
                suffix = sfx
                break

        api_base = ALEXA_API_BASE.get(suffix)
        if not api_base:
            api_base = f"https://alexa.{domain}"

        login._alexa_api_url = api_base.rstrip("/")
        _LOGGER.debug(
            "%s: Falling back to default alexa API base: %s",
            hide_email(login.email),
            login._alexa_api_url,
        )
        return login._alexa_api_url

    @classmethod
    async def _process_response(
        cls, response: ClientResponse, login: AlexaLogin
    ) -> ClientResponse | None:
        """Process a response from _request or static_request.

        Args:
            ClientResponse (response): Response from _request

        Returns:
            None | ClientResponse: Response from server
        """
        if login.stats:
            login.stats["api_calls"] += 1
            _LOGGER.debug("api_calls: %s", login.stats["api_calls"])
        if response.status == 401:
            if login.status:
                login.status["login_successful"] = False
            raise AlexapyLoginError(response.reason)
        if response.status == 429:
            raise AlexapyTooManyRequestsError(response.reason)
        if response.status >= 400:
            _LOGGER.debug("Returning None due to status: %s", response.status)
            return None
        return response

    @backoff.on_exception(
        _min_expo_wait(random.uniform(0.5, 1.5)),
        (AlexapyTooManyRequestsError, AlexapyConnectionError, ClientConnectionError),
        max_time=90,
        max_tries=10,
        jitter=None,
        # factor = 2,
        logger=__name__,
    )
    async def _request(
        self,
        method: str,
        uri: str,
        data: dict[str, str] | None = None,
        query: dict[str, str | int] | None = None,
    ) -> ClientResponse | None:
        async with self._login._oauth_lock:
            if self._login.expires_in and (self._login.expires_in - time.time() < 0):
                _LOGGER.debug(
                    "%s: Detected access token expiration; refreshing",
                    hide_email(self._login.email),
                )
                if (
                    # await self._login.get_tokens()
                    # and
                    await self._login.refresh_access_token()
                    and await self._login.exchange_token_for_cookies()
                    and await self._login.get_csrf()
                ):
                    await self._login.finalize_login()
                else:
                    _LOGGER.debug(
                        "%s: Unable to refresh oauth",
                        hide_email(self._login.email),
                    )
                    self._login.access_token = None
                    self._login.refresh_token = None
                    self._login.expires_in = None
        if method == "get":
            if query and not query.get("_"):
                query["_"] = math.floor(time.time() * 1000)
            elif query is None:
                query = {"_": math.floor(time.time() * 1000)}
        url: URL = URL(self._url + uri).update_query(query)
        _LOGGER.debug(
            "%s: Trying %s: %s : with uri: %s data %s query %s",
            hide_email(self._login.email),
            method,
            url,
            uri,
            data,
            query,
        )
        if self._login.close_requested:
            _LOGGER.debug(
                "%s: Login object has been asked to close; ignoring %s request to %s with %s %s",  # noqa: E501
                hide_email(self._login.email),
                method,
                uri,
                data,
                query,
            )
            raise AlexapyLoginCloseRequested()
        if self._login.status and not self._login.status.get("login_successful"):
            _LOGGER.debug(
                "%s:Login error detected; ignoring %s request to %s with %s %s",
                hide_email(self._login.email),
                method,
                uri,
                data,
                query,
            )
            raise AlexapyLoginError("Login error detected; not contacting API")
        if self._session and self._session.closed:
            raise AlexapyLoginError("Session is closed")
        response = await getattr(self._session, method)(
            url,
            json=data,
            # cookies=self._login._cookies,
            headers=self._login._headers,
            ssl=self._login._ssl,
        )
        _LOGGER.debug(
            "%s: %s: %s returned %s:%s:%s",
            hide_email(self._login.email),
            response.request_info.method,
            response.request_info.url,
            response.status,
            response.reason,
            response.content_type,
        )
        return await self._process_response(response, self._login)

    async def _post_request(
        self,
        uri: str,
        data: dict[str, Any] | None = None,
        query: dict[str, Any] | None = None,
    ) -> ClientResponse | None:
        return await self._request("post", uri, data, query)

    async def _put_request(
        self,
        uri: str,
        data: dict[str, Any] | None = None,
        query: dict[str, Any] | None = None,
    ) -> ClientResponse | None:
        return await self._request("put", uri, data, query)

    async def _get_request(
        self,
        uri: str,
        data: dict[str, Any] | None = None,
        query: dict[str, Any] | None = None,
    ) -> ClientResponse | None:
        return await self._request("get", uri, data, query)

    async def _del_request(
        self,
        uri: str,
        data: dict[str, Any] | None = None,
        query: dict[str, Any] | None = None,
    ) -> ClientResponse | None:
        return await self._request("delete", uri, data, query)

    @staticmethod
    @backoff.on_exception(
        _min_expo_wait(random.uniform(0.8, 1.8)),
        (AlexapyTooManyRequestsError, AlexapyConnectionError, ClientConnectionError),
        max_time=120,
        max_tries=10,
        jitter=None,
        # factor = 2,
        logger=__name__,
    )
    async def _static_request(
        method: str,
        login: AlexaLogin,
        uri: str,
        data: dict[str, str | dict] | None = None,
        additional_headers: dict[str, str] | None = None,
        query: dict[str, str | int] | None = None,
        sub_domain: str = "alexa",
    ) -> ClientResponse | None:
        async with login._oauth_lock:
            if login.expires_in and (login.expires_in - time.time() < 0):
                _LOGGER.debug(
                    "%s: Detected access token expiration; refreshing",
                    hide_email(login.email),
                )
                if (
                    # await login.get_tokens()
                    # and
                    await login.refresh_access_token()
                    and await login.exchange_token_for_cookies()
                    and await login.get_csrf()
                ):
                    await login.finalize_login()
                else:
                    _LOGGER.debug(
                        "%s: Unable to refresh oauth",
                        hide_email(login.email),
                    )
                    login.access_token = None
                    login.refresh_token = None
                    login.expires_in = None
        session = login.session
        url: URL = URL("https://" + sub_domain + "." + login.url + uri).update_query(
            query
        )
        # _LOGGER.debug("%s: %s: Trying static %s: %s : with uri: %s data %s query %s",
        #                hide_email(login.email)
        #                method,
        #                url,
        #                uri,
        #                data,
        #                query,
        #            )
        if login.close_requested:
            _LOGGER.debug(
                "%s: Login object has been asked to close; ignoring %s request to %s with %s %s",  # noqa: E501
                hide_email(login.email),
                method,
                uri,
                data,
                query,
            )
            raise AlexapyLoginCloseRequested()
        if login.status and not login.status.get("login_successful"):
            _LOGGER.debug(
                "%s: Login error detected; ignoring %s request to %s with %s %s",
                hide_email(login.email),
                method,
                uri,
                data,
                query,
            )
            raise AlexapyLoginError("Login error detected; not contacting API")
        if session and session.closed:
            raise AlexapyLoginError("Session is closed")
        headers = login._headers.copy()
        if additional_headers:
            headers.update(additional_headers)
        response = await getattr(session, method)(
            url,
            json=data,
            headers=headers,
            ssl=login._ssl,
        )
        _LOGGER.debug(
            "%s: static %s %s returned %s:%s:%s",
            hide_email(login.email),
            response.request_info.method,
            response.request_info.url,
            response.status,
            response.reason,
            response.content_type,
        )

        # (401 retry handling stays as-is here)
        if response.status == 401 and await login.test_loggedin():
            response = await getattr(session, method)(
                url,
                json=data,
                # cookies=login._cookies,
                headers=login._headers,
                ssl=login._ssl,
            )
            _LOGGER.debug(
                "Error 401, retried once request: %s: static %s: %s returned %s:%s:%s",
                hide_email(login.email),
                response.request_info.method,
                response.request_info.url,
                response.status,
                response.reason,
                response.content_type,
            )

        # ★★★ Error body logger + basic throttling retry ★★★
        if response.status >= 400:
            try:
                body = await response.text()
            except Exception:
                body = "<unable to read body>"

            _LOGGER.debug(
                "HTTP static %s %s returned %s:%s; content_type=%s, body: %s",
                hide_email(login.email),
                response.request_info.url,
                response.status,
                response.reason,
                response.content_type,
                hide_serial(body),
            )

            # Simple, alexa-remote-style throttle handling
            if (
                response.status in (400, 429)
                and isinstance(body, str)
                and any(
                    marker in body
                    for marker in (
                        "Rate exceeded",
                        "ThrottlingException",
                        "Too many requests"
                    )
                )
                and not getattr(login, "_rate_retry_inflight", False)
            ):
                # Choose a delay similar to alexa-remote.js
                delay = random.uniform(10.0, 13.0)
                if "Too many requests" in body:
                    delay += 20.0 + random.uniform(0.0, 30.0)

                _LOGGER.debug(
                    "%s: Throttled (%s); retrying static %s %s once in %.1fs",
                    hide_email(login.email),
                    body,
                    method,
                    url,
                    delay,
                )

                # Prevent concurrent mass retries on the same login
                login._rate_retry_inflight = True
                try:
                    await asyncio.sleep(delay)
                    return await AlexaAPI._static_request(
                        method,
                        login,
                        uri,
                        data=data,
                        additional_headers=additional_headers,
                        query=query,
                        sub_domain=sub_domain,
                    )
                finally:
                    login._rate_retry_inflight = False

        return await AlexaAPI._process_response(response, login)

    @_catch_all_exceptions
    async def run_behavior(
        self,
        node_data,
        queue_delay: float = 1.5,
    ) -> None:
        """Queue node_data for running a behavior in sequence.

        Amazon sequences and routines are based on node_data.

        Args:
            node_data (dict, list of dicts): The node_data to run.
            queue_delay (float, optional): The number of seconds to wait
                                          for commands to queue together.
                                          Defaults to 1.5.
                                          Must be positive.

        """
        sequence_json: dict[Any, Any] = {
            "@type": "com.amazon.alexa.behaviors.model.Sequence",
            "startNode": node_data,
        }
        if queue_delay is None:
            queue_delay = 1.5
        if queue_delay > 0:
            sequence_json["startNode"] = {
                "@type": "com.amazon.alexa.behaviors.model.SerialNode",
                "nodesToExecute": [],
            }
            async with AlexaAPI._sequence_lock[self._login.email]:
                if AlexaAPI._sequence_queue[self._login.email]:
                    last_node = AlexaAPI._sequence_queue[self._login.email][-1]
                    new_node = node_data
                    if node_data and isinstance(node_data, list):
                        new_node = node_data[0]
                    if (
                        last_node.get("operationPayload", {}).get("deviceSerialNumber")
                        and new_node.get("operationPayload", {}).get(
                            "deviceSerialNumber"
                        )
                    ) and last_node.get("operationPayload", {}).get(
                        "deviceSerialNumber"
                    ) != new_node.get("operationPayload", {}).get("deviceSerialNumber"):
                        _LOGGER.debug(
                            "%s: Creating Parallel node",
                            hide_email(self._login.email),
                        )
                        sequence_json["startNode"]["@type"] = (
                            "com.amazon.alexa.behaviors.model.ParallelNode"
                        )
                if isinstance(node_data, list):
                    AlexaAPI._sequence_queue[self._login.email].extend(node_data)
                else:
                    AlexaAPI._sequence_queue[self._login.email].append(node_data)
                items = len(AlexaAPI._sequence_queue[self._login.email])
                old_sequence: list[dict[Any, Any]] = AlexaAPI._sequence_queue[
                    self._login.email
                ]
            await asyncio.sleep(queue_delay)
            async with AlexaAPI._sequence_lock[self._login.email]:
                if (
                    items == len(AlexaAPI._sequence_queue[self._login.email])
                    and old_sequence == AlexaAPI._sequence_queue[self._login.email]
                ):
                    sequence_json["startNode"]["nodesToExecute"].extend(
                        AlexaAPI._sequence_queue[self._login.email]
                    )
                    AlexaAPI._sequence_queue[self._login.email] = []
                    _LOGGER.debug(
                        "%s: Creating sequence for %s items",
                        hide_email(self._login.email),
                        items,
                    )
                else:
                    _LOGGER.debug(
                        "%s: Queue changed while waiting %s seconds",
                        hide_email(self._login.email),
                        queue_delay,
                    )
                    return
        data = {
            "behaviorId": "PREVIEW",
            "sequenceJson": json.dumps(sequence_json),
            "status": "ENABLED",
        }
        _LOGGER.debug(
            "%s: Running behavior with data: %s",
            hide_email(self._login.email),
            json.dumps(data),
        )
        await self._post_request("/api/behaviors/preview", data=data)

    @_catch_all_exceptions
    async def send_sequence(
        self,
        sequence: str,
        customer_id: str | None = None,
        queue_delay: float = 1.5,
        extra: dict[Any, Any] | None = None,
        **kwargs,
    ) -> None:
        """Send sequence command.

        This allows some programmatic control of Echo device using the behaviors
        API and is the basis of play_music, send_announcement, and send_tts.

        Args:
        sequence (string): The Alexa sequence.  Supported list below.
        customer_id (string): CustomerId to use for authorization. When none
                             specified this defaults to the logged in user. Used
                             with households where others may have their own
                             music.
        queue_delay (float, optional): The number of seconds to wait
                                    for commands to queue together.
                                    Defaults to 1.5.
                                    Must be positive.
        extra (Dict): Extra dictionary array; functionality undetermined
        **kwargs : Each named variable must match a recognized Amazon variable
                   within the operationPayload. Please see examples in
                   play_music, send_announcement, and send_tts.
                   Variables with value None are removed from the operationPayload.
                   Variables prefixed with "root_" will be added to the root node instead.

        Supported sequences:
        Alexa.Weather.Play
        Alexa.Traffic.Play
        Alexa.FlashBriefing.Play
        Alexa.GoodMorning.Play
        Alexa.GoodNight.Play
        Alexa.SingASong.Play
        Alexa.TellStory.Play
        Alexa.FunFact.Play
        Alexa.Joke.Play
        Alexa.CleanUp.Play
        Alexa.Music.PlaySearchPhrase
        Alexa.Calendar.PlayTomorrow
        Alexa.Calendar.PlayToday
        Alexa.Calendar.PlayNext
        https://github.com/custom-components/alexa_media_player/wiki#sequence-commands-versions--100

        """  # noqa: E501
        extra = extra or {}
        operation_payload = {
            "deviceType": self._device._device_type,
            "deviceSerialNumber": self._device.device_serial_number,
            "locale": (
                self._device._locale if self._device._locale else DEFAULT_LOCALE
            ),
            "customerId": (
                self._login.customer_id if customer_id is None else customer_id
            ),
        }
        root_node = {}
        if kwargs is not None:
            operation_payload.update(kwargs)
            for key, value in kwargs.items():
                if value is None:  # remove null keys
                    operation_payload.pop(key)
                elif isinstance(value, str) and value.startswith("root_"):
                    operation_payload.pop(key)
                    root_node[key] = value[5:]
            if kwargs.get("devices"):
                operation_payload.pop("deviceType")
                operation_payload.pop("deviceSerialNumber")
        node_data = {
            "@type": "com.amazon.alexa.behaviors.model.OpaquePayloadOperationNode",
            "type": sequence,
            "operationPayload": operation_payload,
        }
        node_data.update(root_node)
        await self.run_behavior(node_data, queue_delay=queue_delay)

    @_catch_all_exceptions
    async def run_skill(
        self,
        skill_id: str,
        customer_id: str | None = None,
        queue_delay: float = 0,
    ) -> None:
        """Run Alexa skill.

        This allows running of defined Alexa skill.

        Args:
            skill_id (string): The full skill id.
            customer_id (string): CustomerId to use for authorization. When none
                             specified this defaults to the logged in user. Used
                             with households where others may have their own
                             music.
            queue_delay (float, optional): The number of seconds to wait
                                        for commands to queue together.
                                        Defaults to 1.5.
                                        Must be positive.

        """
        operation_payload = {
            "targetDevice": {
                "deviceType": self._device._device_type,
                "deviceSerialNumber": self._device.device_serial_number,
            },
            "locale": (
                self._device._locale if self._device._locale else DEFAULT_LOCALE
            ),
            "customerId": (
                self._login.customer_id if customer_id is None else customer_id
            ),
            "connectionRequest": {
                "uri": "connection://AMAZON.Launch/" + skill_id,
                "input": {},
            },
        }
        node_data = {
            "@type": "com.amazon.alexa.behaviors.model.OpaquePayloadOperationNode",
            "type": "Alexa.Operation.SkillConnections.Launch",
            "operationPayload": operation_payload,
        }
        await self.run_behavior(node_data, queue_delay=queue_delay)

    @_catch_all_exceptions
    async def run_custom(
        self,
        text: str,
        customer_id: str | None = None,
        queue_delay: float = 0,
        extra: dict[Any, Any] | None = None,
    ) -> None:
        """Run Alexa skill.

        This allows running exactly what you can say to alexa.

        Args:
            text (string): The full text you want alexa to execute.
            customer_id (string): CustomerId to use for authorization. When none
                             specified this defaults to the logged in user. Used
                             with households where others may have their own
                             music.
            queue_delay (float, optional): The number of seconds to wait
                                        for commands to queue together.
                                        Defaults to 1.5.
                                        Must be positive.
            extra (Dict): Extra dictionary array; functionality undetermined

        """
        extra = extra or {}
        await self.send_sequence(
            "Alexa.TextCommand",
            skillId="amzn1.ask.1p.tellalexa",
            text=text,
            queue_delay=queue_delay,
        )

    @_catch_all_exceptions
    async def run_routine(
        self,
        utterance: str,
        customer_id: str | None = None,
        queue_delay: float = 1.5,
    ) -> None:
        """Run Alexa automation routine.

        This allows running of defined Alexa automation routines.

        Args:
            utterance (string):
                The Alexa routine name or its voice utterance to run the routine.
            customer_id (string):
                CustomerId to use for authorization.
                When none specified this defaults to the logged in user.
                Used with households where others may have their own music.
            queue_delay (float, optional):
                The number of seconds to wait for commands to queue together.
                Defaults to 1.5.
                Must be positive.

        """

        def _populate_device_info(node) -> None:
            """Walk node structure and replace ALEXA_CURRENT_* placeholders."""
            if node is None:
                return

            if isinstance(node, list):
                for item in node:
                    _populate_device_info(item)
                return

            if not isinstance(node, dict):
                return

            if node.get("deviceType") == "ALEXA_CURRENT_DEVICE_TYPE":
                node["deviceType"] = self._device._device_type
            if node.get("deviceSerialNumber") == "ALEXA_CURRENT_DSN":
                node["deviceSerialNumber"] = self._device.device_serial_number
            if node.get("locale") == "ALEXA_CURRENT_LOCALE":
                node["locale"] = self._device._locale or DEFAULT_LOCALE

            for v in node.values():
                if isinstance(v, (dict, list)):
                    _populate_device_info(v)

        # Fetch automations (routines)
        if not (automations := await AlexaAPI.get_automations(self._login)):
            return

        utterance_cf = utterance.casefold()
        automation_id = None
        sequence = None

        for automation in automations:
            if not isinstance(automation, dict):
                continue

            # 1) Name match first (covers device-trigger routines)
            if (
                isinstance(name := automation.get("name"), str)
                and name.casefold() == utterance_cf
            ):
                automation_id = automation.get("automationId")
                seq = automation.get("sequence")
                sequence = seq if isinstance(seq, dict) else None
                break

            # 2) Fallback: spoken utterance match (CustomUtterance only)
            if not (
                isinstance(triggers := automation.get("triggers"), list)
                and triggers
                and isinstance(triggers[0], dict)
                and triggers[0].get("type") == "CustomUtterance"
                and isinstance(payload := triggers[0].get("payload"), dict)
            ):
                continue

            if (
                isinstance(u := payload.get("utterance"), str)
                and u.casefold() == utterance_cf
            ) or (
                isinstance(utterances := payload.get("utterances"), list)
                and any(
                    isinstance(x, str) and x.casefold() == utterance_cf
                    for x in utterances
                )
            ):
                automation_id = automation.get("automationId")
                seq = automation.get("sequence")
                sequence = seq if isinstance(seq, dict) else None
                break

        if automation_id is None or sequence is None:
            _LOGGER.debug(
                "%s: No routine found for %s",
                hide_email(self._login.email),
                utterance,
            )
            return

        if not isinstance(start_node := sequence.get("startNode"), dict):
            _LOGGER.debug(
                "%s: Routine %s found for %s, but sequence.startNode is "
                "missing/invalid",
                hide_email(self._login.email),
                automation_id,
                utterance,
            )
            return

        _populate_device_info(start_node)

        nodes = start_node.get("nodesToExecute")
        await self.run_behavior(
            nodes if isinstance(nodes, list) else start_node,
            queue_delay=queue_delay,
        )

    @_catch_all_exceptions
    async def play_music(
        self,
        provider_id: str,
        search_phrase: str,
        customer_id: str | None = None,
        timer: int | None = None,
        queue_delay: float = 1.5,
        extra: dict[Any, Any] | None = None,
    ) -> None:
        """Play music based on search.

        Args:
            provider_id (str): Amazon music provider.
            search_phrase (str): Phrase to be searched for
            customer_id (Optional[str], optional): CustomerId to use for authorization. When none
                             specified this defaults to the logged in user. Used
                             with households where others may have their own
                             music.
            timer (Optional[int]): Number of seconds to play before stopping.
            queue_delay (float, optional): The number of seconds to wait
                                   for commands to queue together.
                                   Must be positive. Defaults to 1.5.
            extra (Dict): Extra dictionary array; functionality undetermined

        """  # noqa: E501
        extra = extra or {}
        customer_id = self._login.customer_id if customer_id is None else customer_id
        if timer:
            await self.send_sequence(
                "Alexa.Music.PlaySearchPhrase",
                customer_id=customer_id,
                searchPhrase=search_phrase,
                sanitizedSearchPhrase=search_phrase,
                musicProviderId=provider_id,
                waitTimeInSeconds=timer,
                queue_delay=queue_delay,
            )
        else:
            await self.send_sequence(
                "Alexa.Music.PlaySearchPhrase",
                customer_id=customer_id,
                searchPhrase=search_phrase,
                sanitizedSearchPhrase=search_phrase,
                musicProviderId=provider_id,
                queue_delay=queue_delay,
            )

    @_catch_all_exceptions
    async def play_sound(
        self,
        sound_string_id: str,
        customer_id: str | None = None,
        queue_delay: float = 1.5,
        extra: dict[Any, Any] | None = None,
    ) -> None:
        """Play Alexa sound."""
        extra = extra or {}
        await self.send_sequence(
            "Alexa.Sound",
            customer_id=self._login.customer_id if customer_id is None else customer_id,
            soundStringId=sound_string_id,
            skillId="amzn1.ask.1p.sound",
            queue_delay=queue_delay,
        )

    @_catch_all_exceptions
    async def stop(
        self,
        customer_id: str | None = None,
        queue_delay: float = 1.5,
        all_devices: bool = False,
    ) -> None:
        """Stop device playback.

        Keyword Arguments:
            customer_id {str} -- CustomerId issuing command (default: {None})
            queue_delay {float} -- The number of seconds to wait
                                   for commands to queue together.
                                   Must be positive.
                                   (default: {1.5})
            all_devices {bool} -- Whether all devices should be stopped (default: {False})

        """  # noqa: E501
        kwargs = {}

        if all_devices:
            kwargs["devices"] = (
                {
                    "deviceType": "ALEXA_ALL_DEVICE_TYPE",
                    "deviceSerialNumber": "ALEXA_ALL_DSN",
                },
            )
        else:
            kwargs["devices"] = [
                {
                    "deviceSerialNumber": self._device.device_serial_number,
                    "deviceType": self._device._device_type,
                },
            ]

        await self.send_sequence(
            "Alexa.DeviceControls.Stop",
            skillId="amzn1.ask.1p.alexadevicecontrols",
            customer_id=self._login.customer_id if customer_id is None else customer_id,
            queue_delay=queue_delay,
            **kwargs,
        )

    def process_targets(self, targets: list[str] | None = None) -> list[dict[str, str]]:
        """Process targets list to generate list of devices.

        Keyword Arguments
            targets {Optional[List[str]]} -- List of serial numbers
                (default: {[]})

        Returns
            List[Dict[str, str] -- List of device dicts

        """
        targets = targets or []
        devices = []
        if self._device._device_family == "WHA":
            # Build group of devices based off _cluster_members
            for dev in AlexaAPI.devices[self._login.email]:
                if dev["serialNumber"] in self._device._cluster_members:
                    devices.append(
                        {
                            "deviceSerialNumber": dev["serialNumber"],
                            "deviceTypeId": dev["deviceType"],
                        }
                    )
        elif targets and isinstance(targets, list):
            for dev in AlexaAPI.devices[self._login.email]:
                if dev["serialNumber"] in targets or dev["accountName"] in targets:
                    devices.append(
                        {
                            "deviceSerialNumber": dev["serialNumber"],
                            "deviceTypeId": dev["deviceType"],
                        }
                    )
        else:
            devices.append(
                {
                    "deviceSerialNumber": self._device.device_serial_number,
                    "deviceTypeId": self._device._device_type,
                }
            )
        return devices

    @_catch_all_exceptions
    async def send_tts(
        self,
        message: str,
        customer_id: str | None = None,
        targets: list[str] | None = None,
        queue_delay: float = 1.5,
    ) -> None:
        """Send message for TTS at speaker.

        This is the old method which used Alexa Simon Says which did not work
        for WHA. This will not beep prior to sending. send_announcement
        should be used instead.

        Args:
        message (string): The message to speak. For canned messages, the message
                            must start with `alexa.cannedtts.speak` as discovered
                            in the routines.
        customer_id (string): CustomerId to use for authorization. When none
                             specified this defaults to the logged in user. Used
                             with households where others may have their own
                             music.
        targets (list(string)): WARNING: This is currently non functional due
                                to Alexa's API and is only included for future
                                proofing.
                                List of serialNumber or accountName to send the
                                tts to. Only those in this AlexaAPI
                                account will be searched. If None, announce
                                will be self.
        queue_delay (float, optional): The number of seconds to wait
                                          for commands to queue together.
                                          Defaults to 1.5.
                                          Must be positive.

        """
        if message.startswith("alexa.cannedtts.speak"):
            await self.send_sequence(
                "Alexa.CannedTts.Speak",
                customer_id=(
                    self._login.customer_id if customer_id is None else customer_id
                ),
                cannedTtsStringId=message,
                skillId="amzn1.ask.1p.saysomething",
                queue_delay=queue_delay,
            )
        else:
            target = {
                "customerId": (
                    self._login.customer_id if customer_id is None else customer_id
                ),
                "devices": self.process_targets(targets),
            }
            await self.send_sequence(
                "Alexa.Speak",
                customer_id=(
                    self._login.customer_id if customer_id is None else customer_id
                ),
                textToSpeak=message,
                target=target,
                skillId="amzn1.ask.1p.saysomething",
                queue_delay=queue_delay,
            )

    @_catch_all_exceptions
    async def send_announcement(
        self,
        message: str,
        method: str = "all",
        title: str = "Announcement",
        customer_id: str | None = None,
        targets: list[str] | None = None,
        queue_delay: float = 1.5,
        extra: dict[Any, Any] | None = None,
    ) -> None:
        """Send announcement to Alexa devices.

        This uses the AlexaAnnouncement and allows visual display on the Show.
        It will beep prior to speaking.

        Args:
        message (string): The message to speak or display.
        method (string): speak, show, or all
        title (string): title to display on Echo show
        customer_id (string): CustomerId to use for authorization. When none
                             specified this defaults to the logged in user. Used
                             with households where others may have their own
                             music.
        targets (list(string)): List of serialNumber or accountName to send the
                                announcement to. Only those in this AlexaAPI
                                account will be searched. If None, announce
                                will be self.
        queue_delay (float, optional): The number of seconds to wait
                                        for commands to queue together.
                                        Defaults to 1.5.
                                        Must be positive.
        extra (Dict): Extra dictionary array; functionality undetermined

        """
        extra = extra or {}
        display = (
            {"title": "", "body": ""}
            if method.lower() == "speak"
            else {"title": title, "body": message}
        )
        speak = (
            {"type": "text", "value": ""}
            if method.lower() == "show"
            else {"type": "text", "value": message}
        )
        content = [
            {
                "locale": (
                    self._device._locale if self._device._locale else DEFAULT_LOCALE
                ),
                "display": display,
                "speak": speak,
            }
        ]
        target = {
            "customerId": (
                self._login.customer_id if customer_id is None else customer_id
            ),
            "devices": self.process_targets(targets),
        }
        await self.send_sequence(
            "AlexaAnnouncement",
            customer_id=self._login.customer_id if customer_id is None else customer_id,
            expireAfter="PT5S",
            content=content,
            target=target,
            skillId="amzn1.ask.1p.routines.messaging",
            queue_delay=queue_delay,
        )

    @_catch_all_exceptions
    async def send_mobilepush(
        self,
        message: str,
        title: str = "AlexaAPI Message",
        customer_id: str | None = None,
        queue_delay: float = 1.5,
        extra: dict[Any, Any] | None = None,
    ) -> None:
        """Send mobile push to Alexa app.

        Push a message to mobile devices with the Alexa App. This probably
        should be a static method.

        Args:
        message (string): The message to push to the mobile device.
        title (string): Title for push notification
        customer_id (string): CustomerId to use for sending. When none
                              specified this defaults to the logged in user.
        queue_delay (float, optional): The number of seconds to wait
                                        for commands to queue together.
                                        Defaults to 1.5.
                                        Must be positive.
        extra (Dict): Extra dictionary array; functionality undetermined

        """
        extra = extra or {}
        await self.send_sequence(
            "Alexa.Notifications.SendMobilePush",
            customer_id=(
                self._login.customer_id if customer_id is None else customer_id
            ),
            notificationMessage=message,
            alexaUrl="#v2/behaviors",
            title=title,
            skillId="amzn1.ask.1p.routines.messaging",
            queue_delay=queue_delay,
        )

    @_catch_all_exceptions
    async def send_dropin_notification(
        self,
        message: str,
        title: str = "AlexaAPI Dropin Notification",
        customer_id: str | None = None,
        queue_delay: float = 1.5,
        extra: dict[Any, Any] | None = None,
    ) -> None:
        """Send dropin notification to Alexa app for Alexa device.

        Push a message to mobile devices with the Alexa App. This can spawn a
        notification to drop in on a specific device.

        Args:
        message (string): The message to push to the mobile device.
        title (string): Title for push notification
        customer_id (string): CustomerId to use for sending. When none
                              specified this defaults to the logged in user.
        queue_delay (float, optional): The number of seconds to wait
                                        for commands to queue together.
                                        Defaults to 1.5.
                                        Must be positive.
        extra (Dict): Extra dictionary array; functionality undetermined

        """
        extra = extra or {}
        await self.send_sequence(
            "Alexa.Notifications.DropIn",
            customer_id=(
                self._login.customer_id if customer_id is None else customer_id
            ),
            notificationMessage=message,
            alexaUrl="#v2/comms/conversation-list?showDropInDialog=true",
            title=title,
            skillId="root_amzn1.ask.1p.action.dropin",
            queue_delay=queue_delay,
            deviceType=None,
            deviceSerialNumber=None,
            locale=None,
        )

    async def set_media(self, data: dict[str, Any]) -> None:
        """Select the media player."""
        await self._post_request(
            "/api/np/command",
            data=data,
            query={
                "deviceSerialNumber": self._device.device_serial_number,
                "deviceType": self._device._device_type,
            },
        )

    @_catch_all_exceptions
    async def previous(self) -> None:
        """Play previous."""
        await self.set_media({"type": "PreviousCommand"})

    @_catch_all_exceptions
    async def next(self) -> None:
        """Play next."""
        await self.set_media({"type": "NextCommand"})

    @_catch_all_exceptions
    async def pause(self) -> None:
        """Pause."""
        await self.set_media({"type": "PauseCommand"})

    @_catch_all_exceptions
    async def play(self) -> None:
        """Play."""
        await self.set_media({"type": "PlayCommand"})

    @_catch_all_exceptions
    async def forward(self) -> None:
        """Fastforward."""
        await self.set_media({"type": "ForwardCommand"})

    @_catch_all_exceptions
    async def rewind(self) -> None:
        """Rewind."""
        await self.set_media({"type": "RewindCommand"})

    @_catch_all_exceptions
    async def set_volume(
        self,
        volume: float,
        customer_id: str | None = None,
        queue_delay: float = 1.5,
    ) -> None:
        """Set volume.

        Args:
        volume (float): The volume between 0 and 1.
        customer_id (string): CustomerId to use for sending. When none
                              specified this defaults to the logged in user.
        queue_delay (float, optional): The number of seconds to wait
                                        for commands to queue together.
                                        Defaults to 1.5.
                                        Must be positive.

        """
        await self.send_sequence(
            "Alexa.DeviceControls.Volume",
            customer_id=(
                self._login.customer_id if customer_id is None else customer_id
            ),
            value=volume * 100,
            queue_delay=queue_delay,
        )

    @_catch_all_exceptions
    async def shuffle(self, setting: bool) -> None:
        """Shuffle.

        setting (string) : true or false
        """
        await self.set_media({"type": "ShuffleCommand", "shuffle": setting})

    @_catch_all_exceptions
    async def repeat(self, setting: bool) -> None:
        """Repeat.

        setting (string) : true or false
        """
        await self.set_media({"type": "RepeatCommand", "repeat": setting})

    @_catch_all_exceptions
    async def get_state(self) -> dict[str, Any] | None:
        """Get playing state."""
        response = await self._get_request(
            "/api/np/player",
            query={
                "deviceSerialNumber": self._device.device_serial_number,
                "deviceType": self._device._device_type,
                "screenWidth": 2560,
            },
        )
        response_json, *_ = await get_json_value(response, None, dict)
        return response_json

    @_catch_all_exceptions
    async def get_wifi_details(self) -> dict[str, Any] | None:
        """Get wifi details."""
        response = await self._get_request(
            "/api/device-wifi-details",
            query={
                "deviceSerialNumber": self._device.device_serial_number,
                "deviceType": self._device._device_type,
            },
        )
        response_json, *_ = await get_json_value(response, None, dict)
        return response_json

    @_catch_all_exceptions
    async def set_dnd_state(self, state: bool) -> None:
        """Set Do Not Disturb state.

        Args:
        state (boolean): true or false

        Returns json

        """
        data = {
            "deviceSerialNumber": self._device.device_serial_number,
            "deviceType": self._device._device_type,
            "enabled": state,
        }
        _LOGGER.debug(
            "%s: Setting DND state: %s data: %s",
            hide_email(self._login.email),
            state,
            json.dumps(data),
        )
        response = await self._put_request("/api/dnd/status", data=data)
        response_json, *_ = await get_json_value(response, None, dict)
        success = data == response_json
        _LOGGER.debug(
            "%s: Success: %s Response: %s",
            hide_email(self._login.email),
            success,
            response_json,
        )
        return success

    @staticmethod
    @_catch_all_exceptions
    async def get_bluetooth(login) -> dict[str, Any] | None:
        """Get paired bluetooth devices."""
        response = await AlexaAPI._static_request(
            "get", login, "/api/bluetooth", query={"cached": "false"}
        )
        response_json, *_ = await get_json_value(response, None, dict)
        return response_json

    @_catch_all_exceptions
    async def set_bluetooth(self, mac: str) -> None:
        """Pair with bluetooth device with mac address."""
        await self._post_request(
            "/api/bluetooth/pair-sink/"
            + self._device._device_type
            + "/"
            + self._device.device_serial_number,
            data={"bluetoothDeviceAddress": mac},
        )

    @_catch_all_exceptions
    async def disconnect_bluetooth(self) -> None:
        """Disconnect all bluetooth devices."""
        await self._post_request(
            "/api/bluetooth/disconnect-sink/"
            + self._device._device_type
            + "/"
            + self._device.device_serial_number,
            data=None,
        )

    @staticmethod
    @_catch_all_exceptions
    async def get_devices(login: AlexaLogin) -> list[dict[str, Any]] | None:
        """Identify all Alexa devices."""
        response = await AlexaAPI._static_request(
            "get", login, "/api/devices-v2/device", query=None
        )
        devices, *_ = await get_json_value(response, "devices", list)
        AlexaAPI.devices[login.email] = (
            devices if devices else AlexaAPI.devices[login.email]
        )
        return AlexaAPI.devices[login.email]

    @staticmethod
    @_catch_all_exceptions
    async def get_wake_words(login: AlexaLogin) -> list[dict[str, Any]] | None:
        """Get the wake words for the devices."""
        response = await AlexaAPI._static_request(
            "get", login, "/api/wake-word", query={"cached": "true"}
        )
        wakeWords, *_ = await get_json_value(response, "wakeWords", list)
        AlexaAPI.wake_words[login.email] = (
            wakeWords if wakeWords else AlexaAPI.wake_words[login.email]
        )
        return AlexaAPI.wake_words[login.email]

    @staticmethod
    @_catch_all_exceptions
    async def find_wake_word(login: AlexaLogin, serial: str) -> str | None:
        """Find the wake word associated to a device."""
        wake_words = (
            AlexaAPI.wake_words[login.email]
            if login.email in AlexaAPI.wake_words
            else await AlexaAPI.get_wake_words(login)
        )
        if wake_words:
            found = next(
                filter(
                    lambda wake_word: serial == wake_word["deviceSerialNumber"],
                    wake_words,
                ),
                None,
            )
            if found is not None:
                return found["wakeWord"].lower()
        return None

    @staticmethod
    @_catch_all_exceptions
    async def get_authentication(login: AlexaLogin) -> dict[str, Any] | None:
        """Get authentication json."""
        response = await AlexaAPI._static_request(
            "get",
            login,
            f"/api/users/me?platform=ios&version={CALL_VERSION}",
            query=None,
        )
        json_resp, *_ = await get_json_value(response, None, dict)
        if json_resp:
            return {
                "authenticated": True,
                "canAccessPrimeMusicContent": json_resp.get(
                    "canAccessPrimeMusicContent", True
                ),
                "customerEmail": json_resp.get("email"),
                "customerId": json_resp.get("id"),
                "customerName": json_resp.get("fullName"),
            }
        return None

    @staticmethod
    @_catch_all_exceptions
    async def get_customer_history_records(
        login: AlexaLogin,
        start_time: int | None = None,
        end_time: int | None = None,
        max_record_size: int | None = 1,
    ) -> list[dict[str, Any]] | None:
        """Get customer history records."""
        start_time = (
            int((time.time() - 24 * 3600) * 1000) if start_time is None else start_time
        )
        end_time = (
            int((time.time() + 24 * 3600) * 1000) if end_time is None else end_time
        )
        extra_headers = {
            "referer": f"https://www.{login.url}",
            "anti-csrftoken-a2z": login.csrf_token,
        }
        # The anti-csrf token expires after 24 hours
        if login.csrf_token is None or (
            int(time.time()) - login.csrf_token_created_at > 60 * 60 * 24
        ):
            extra_headers["anti-csrftoken-a2z"] = await login.get_csrf_token()

        response = await AlexaAPI._static_request(
            "post",
            login,
            "/alexa-privacy/apd/rvh/customer-history-records-v2",
            data={"previousRequestToken": None},
            additional_headers=extra_headers,
            query={
                "startTime": start_time,
                "endTime": end_time,
                "recordType": "VOICE_HISTORY",
                "maxRecordSize": max_record_size,
            },
            sub_domain="www",
        )
        if response is None:
            return None
        ret = []
        customerHistoryRecords, valid = await get_json_value(
            response, "customerHistoryRecords", (list, NoneType)
        )
        if customerHistoryRecords is None:
            return ret if valid else None

        for record in customerHistoryRecords:
            o = {}
            conv_parts = {}
            if record["voiceHistoryRecordItems"]:
                for item in record["voiceHistoryRecordItems"]:
                    conv_parts[item["recordItemType"]] = conv_parts.get(
                        item["recordItemType"], []
                    )
                    conv_parts[item["recordItemType"]].append(item)

                o["conversionDetails"] = conv_parts

            record_key = record["recordKey"].split("#")
            o["deviceType"] = record_key[2] if record_key[2] else None
            o["creationTimestamp"] = record.get("timestamp", None)
            o["deviceSerialNumber"] = record_key[3]
            o["description"] = {"summary": ""}
            o["utteranceType"] = record["utteranceType"]
            wake_word = await AlexaAPI.find_wake_word(login, record_key[3])
            if (
                "CUSTOMER_TRANSCRIPT" in conv_parts
                or "ASR_REPLACEMENT_TEXT" in conv_parts
            ):
                if "CUSTOMER_TRANSCRIPT" in conv_parts:
                    for trans in conv_parts["CUSTOMER_TRANSCRIPT"]:
                        text = trans["transcriptText"]
                        if wake_word and text.startswith(wake_word):
                            text = text[len(wake_word) :].strip()
                        elif text.startswith("alexa"):
                            text = text[5:].strip()
                        o["description"]["summary"] += text + ", "

                if "ASR_REPLACEMENT_TEXT" in conv_parts:
                    for trans in conv_parts["ASR_REPLACEMENT_TEXT"]:
                        text = trans["transcriptText"]
                        if wake_word and text.startswith(wake_word):
                            text = text[len(wake_word) :].strip()
                        elif text.startswith("alexa"):
                            text = text[5:].strip()
                        o["description"]["summary"] += text + ", "
                o["description"]["summary"] = o["description"]["summary"][
                    0 : len(o["description"]["summary"]) - 2
                ].strip()

                o["alexaResponse"] = ""
                if (
                    "ALEXA_RESPONSE" in conv_parts
                    or "TTS_REPLACEMENT_TEXT" in conv_parts
                ):
                    if "ALEXA_RESPONSE" in conv_parts:
                        for trans in conv_parts["ALEXA_RESPONSE"]:
                            o["alexaResponse"] += trans["transcriptText"] + ", "

                    if "TTS_REPLACEMENT_TEXT" in conv_parts:
                        for trans in conv_parts["TTS_REPLACEMENT_TEXT"]:
                            o["alexaResponse"] += trans["transcriptText"] + ", "
                    o["alexaResponse"] = o["alexaResponse"][
                        0 : len(o["alexaResponse"]) - 2
                    ].strip()

            ret.append(o)
        return ret

    @staticmethod
    @_catch_all_exceptions
    async def get_activities(
        login: AlexaLogin, items: int = 10
    ) -> dict[str, Any] | None:
        """Get activities json."""
        response = await AlexaAPI._static_request(
            "get",
            login,
            "/api/activities",
            query={"startTime": "", "size": items, "offset": 1},
        )
        result, *_ = await get_json_value(response, "activities", dict)
        return result

    @staticmethod
    @_catch_all_exceptions
    async def get_device_preferences(login: AlexaLogin) -> dict[str, Any] | None:
        """Identify all Alexa device preferences."""
        response = await AlexaAPI._static_request(
            "get", login, "/api/device-preferences", query={}
        )
        response_json, *_ = await get_json_value(response, None, dict)
        return response_json

    @staticmethod
    @_catch_all_exceptions
    async def get_automations(
        login: AlexaLogin, items: int = 1000
    ) -> list[dict[str, Any]] | None:
        """Identify all Alexa automations."""
        response = await AlexaAPI._static_request(
            "get", login, "/api/behaviors/v2/automations", query={"limit": items}
        )
        response_json, *_ = await get_json_value(response, None, list)
        return response_json

    @staticmethod
    @_catch_all_exceptions
    async def get_last_device_serial(
        login: AlexaLogin, items: int = 10
    ) -> dict[str, Any] | None:
        """Identify the last device's serial number and last summary.

        This will search the [last items] activity records and find the latest
        entry where Echo successfully responded.
        """
        response = await AlexaAPI.get_customer_history_records(
            login, max_record_size=items
        )
        if response is not None:
            for last_activity in response:
                summary = ""
                # Ignore empty description and summary
                # Ignore utterance type DEVICE_ARBITRATION
                if (
                    last_activity["description"]
                    and last_activity["description"]["summary"]
                    and last_activity["utteranceType"] != "DEVICE_ARBITRATION"
                ):
                    with contextlib.suppress(AttributeError, JSONDecodeError):
                        summary = last_activity["description"]["summary"]
                    return {
                        "serialNumber": (last_activity["deviceSerialNumber"]),
                        "timestamp": last_activity["creationTimestamp"],
                        "summary": summary,
                    }

        return None

    @_catch_all_exceptions
    async def set_guard_state(
        self, entity_id: str, state: str, queue_delay: float = 1.5
    ) -> None:
        """Set Guard state.

        Args:
        entity_id (str): numeric ending of applianceId of RedRock Panel
        state (str): AWAY, HOME
        queue_delay (float, optional): The number of seconds to wait
                                        for commands to queue together.
                                        Defaults to 1.5.
                                        Must be positive.
        Returns json

        """
        _LOGGER.debug(
            "%s: Setting Guard state: %s ", hide_email(self._login.email), state
        )

        await self.send_sequence(
            "controlGuardState",
            target=entity_id,
            operationId="controlGuardState",
            state=state,
            skillId="amzn1.ask.skill.f71a9b50-e99a-4669-a226-d50ebb5e0830",
            queue_delay=queue_delay,
        )

    @staticmethod
    @_catch_all_exceptions
    async def get_guard_state(
        login: AlexaLogin, entity_id: str
    ) -> dict[str, Any] | None:
        """Get state of Alexa guard.

        Args:
        login (AlexaLogin): Successfully logged in AlexaLogin
        entity_id (str): applianceId of RedRock Panel

        Returns json

        """
        return await AlexaAPI.get_entity_state(login, appliance_ids=[entity_id])

    @staticmethod
    @_catch_all_exceptions
    async def get_entity_state(
        login: AlexaLogin,
        entity_ids: list[str] | None = None,
        appliance_ids: list[str] | None = None,
    ) -> dict[str, Any] | None:
        """Get the current state of multiple appliances.

        Note that this can take both entity_ids and appliance_ids.
        If you have both pieces of data available, prefer the entity id. A single entity might have multiple
        appliance ids. Its easier to ensure you don't miss data by just providing entity id instead.

        Args:
        login (AlexaLogin): Successfully logged in AlexaLogin
        entity_ids (List[str]): The list of entities you want information about.
        appliance_ids: (List[str]): The list of appliances you want information about.

        Returns json

        """  # noqa: E501
        state_requests = []
        if entity_ids is not None:
            for entity_id in entity_ids:
                state_requests.append({"entityId": entity_id, "entityType": "ENTITY"})
        if appliance_ids is not None:
            for appliance_id in appliance_ids:
                state_requests.append(
                    {"entityId": appliance_id, "entityType": "APPLIANCE"}
                )
        data = {"stateRequests": state_requests}
        response = await AlexaAPI._static_request(
            "post", login, "/api/phoenix/state", data=data
        )
        response_json, *_ = await get_json_value(response, None, dict)
        _LOGGER.debug(
            "%s: get_entity_state response: %s", hide_email(login.email), response_json
        )
        return response_json

    @staticmethod
    @_catch_all_exceptions
    async def static_set_guard_state(
        login: AlexaLogin, entity_id: str, state: str
    ) -> dict[str, Any] | None:
        """Set state of Alexa guard.

        Args:
        login (AlexaLogin): Successfully logged in AlexaLogin
        entity_id (str): entityId of RedRock Panel
        state (str): ARMED_AWAY, ARMED_STAY

        Returns json

        """
        parameters = {"action": "controlSecurityPanel", "armState": state}
        data = {
            "controlRequests": [
                {
                    "entityId": entity_id,
                    "entityType": "APPLIANCE",
                    "parameters": parameters,
                }
            ]
        }
        response = await AlexaAPI._static_request(
            "put", login, "/api/phoenix/state", data=data
        )
        response_json, *_ = await get_json_value(response, None, dict)
        _LOGGER.debug(
            "%s: set_guard_state response: %s for data: %s ",
            hide_email(login.email),
            response_json,
            json.dumps(data),
        )
        return response_json

    @staticmethod
    @_catch_all_exceptions
    async def set_light_state(
        login: AlexaLogin,
        entity_id: str,
        power_on: bool = True,
        brightness: int | None = None,
        color_name: str | None = None,
        color_temperature_name: str | None = None,
    ) -> dict[str, Any] | None:
        """Set state of a light.

        Args:
        login (AlexaLogin): Successfully logged in AlexaLogin
        entity_id (str): Entity ID of The light. Not the Application ID.
        power_on (bool): Should the light be on or off.
        brightness (Optional[int]): 0-100 or None to leave as is
        color_name (Optional[str]): The name of a color that Alexa supports in snake case.
        color_temperature_name (Optional[str]): The name of a color temperature name that Alexa supports in snake case.

        Returns json

        """  # noqa: E501
        control_requests = [
            {
                "entityId": entity_id,
                "entityType": "ENTITY",
                "parameters": {"action": "turnOn" if power_on else "turnOff"},
            }
        ]

        if brightness is not None and 0 <= brightness <= 100:
            control_requests.append(
                {
                    "entityId": entity_id,
                    "entityType": "ENTITY",
                    "parameters": {
                        "action": "setBrightness",
                        "brightness": str(brightness),
                    },
                }
            )

        if color_name is not None:
            control_requests.append(
                {
                    "entityId": entity_id,
                    "entityType": "ENTITY",
                    "parameters": {"action": "setColor", "colorName": color_name},
                }
            )

        if color_temperature_name is not None:
            control_requests.append(
                {
                    "entityId": entity_id,
                    "entityType": "ENTITY",
                    "parameters": {
                        "action": "setColorTemperature",
                        "colorTemperatureName": color_temperature_name,
                    },
                }
            )

        data = {"controlRequests": control_requests}

        response = await AlexaAPI._static_request(
            "put", login, "/api/phoenix/state", data=data
        )
        response_json, *_ = await get_json_value(response, None, dict)
        _LOGGER.debug(
            "%s: set_light_state response: %s for data: %s ",
            hide_email(login.email),
            response_json,
            json.dumps(data),
        )
        return response_json

    @staticmethod
    @_catch_all_exceptions
    async def get_devices_gql(login: AlexaLogin) -> list[dict[str, Any]] | None:
        """Get devices of the Alexa network.

        Args:
        login (AlexaLogin): Successfully logged in AlexaLogin

        Returns json

        """
        response = await AlexaAPI._static_request(
            "post", login, "/nexus/v1/graphql", data={"query": GQL_SMARTHOME_QUERY}
        )
        # _LOGGER.debug("%s: Response: %s", hide_email(login.email),
        #               await response.json(content_type=None))
        result, *_ = await get_json_value(response, "data.endpoints.items", list)
        return result

    @staticmethod
    @_catch_all_exceptions
    async def get_network_details(login: AlexaLogin) -> list[dict[str, Any]] | None:
        """Get the network of devices that Alexa is aware of.

        Args:
        login: (AlexaLogin): Successfully logged in AlexaLogin

        Returns json
        """
        network_detail = await AlexaAPI.get_devices_gql(login)
        _LOGGER.debug(
            "%s: get_devices_gql response: %s",
            hide_email(login.email),
            network_detail,
        )
        if network_detail is None:
            return None
        details = []
        for el in network_detail:
            if el.get("legacyAppliance", None) is None:
                return None
            details.append(el["legacyAppliance"])
        return details

    @staticmethod
    def _should_skip_notifications(email: str, now: float) -> bool:
        """Return True if we should skip notifications due to cooldown."""
        if not hasattr(AlexaAPI, "_notif_last_call"):
            AlexaAPI._notif_last_call = {}

        last = AlexaAPI._notif_last_call.get(email, 0.0)
        cooldown = 10
        if now - last < cooldown:
            _LOGGER.debug(
                "%s: Skipping alexapy get_notifications; last %.1fs ago (cooldown %ss)",
                hide_email(email),
                now - last,
                cooldown,
            )
            return True

        AlexaAPI._notif_last_call[email] = now
        return False

    @staticmethod
    def _build_notifications_headers(login: AlexaLogin) -> dict[str, str]:
        """Build headers for the notifications request."""
        base_headers = getattr(login, "_headers", {}) or {}
        cookie = base_headers.get("Cookie") or base_headers.get("cookie")
        csrf = (
            base_headers.get("anti-csrftoken-a2z")
            or base_headers.get("csrf")
            or base_headers.get("x-amzn-csrf")
        )

        headers: dict[str, str] = {}
        if cookie:
            headers["Cookie"] = cookie
        if csrf:
            if "anti-csrftoken-a2z" in base_headers:
                headers["anti-csrftoken-a2z"] = csrf
            else:
                headers["csrf"] = csrf

        headers["User-Agent"] = API_USER_AGENT
        headers["Accept-Language"] = DEFAULT_ACCEPT_LANGUAGE
        headers["Accept"] = "application/json"
        headers["Connection"] = "keep-alive"

        referer = base_headers.get("Referer") or base_headers.get("referer")
        if referer:
            headers["Referer"] = referer

        return headers

    @staticmethod
    @_catch_all_exceptions
    async def get_notifications(login: AlexaLogin) -> list[dict[str, Any]] | None:
        """Get Alexa notifications using discovered API base."""
        email = login.email
        now = time.time()

        if AlexaAPI._should_skip_notifications(email, now):
            return None

        session = login.session
        if session.closed:
            raise AlexapyLoginError("Session is closed")

        # 🔹 Use discovered base instead of hard-coded NA
        api_base = await AlexaAPI._get_alexa_api_base(login)
        ts = int(time.time() * 1000)
        url = URL(api_base).with_path("/api/notifications").update_query(
            {"cached": "true", "_": str(ts)}
        )

        headers = AlexaAPI._build_notifications_headers(login)

        try:
            resp = await session.get(url, headers=headers, ssl=login._ssl)
        except ClientConnectionError as exc:
            raise AlexapyConnectionError(str(exc)) from exc

        body = await resp.text()
        _LOGGER.debug(
            "%s: notifications GET %s returned %s:%s:%s",
            hide_email(email),
            url,
            resp.status,
            resp.reason,
            resp.content_type,
        )
        try:
            parsed = json.loads(body)
            count = len(parsed.get("notifications", []))
            summary = f"{count} notifications"
        except Exception:
            summary = "<non-JSON body>"

        _LOGGER.debug(
            "%s: notifications GET JSON summary: %s",
            hide_email(email),
            summary,
        )

        if resp.status == 400 and "Rate exceeded" in body:
            _LOGGER.debug(
                "%s: notifications throttled (%s); not retrying this call",
                hide_email(email),
                hide_serial(body),
            )
            return None

        if resp.status == 401:
            login.status["login_successful"] = False
            raise AlexapyLoginError(resp.reason)

        if resp.status >= 400:
            _LOGGER.debug(
                "%s: notifications returning None due to status %s; body: %s",
                hide_email(email),
                resp.status,
                hide_serial(body),
            )
            return None

        try:
            data = json.loads(body)
        except ValueError:
            _LOGGER.debug(
                "%s: notifications invalid JSON body: %s",
                hide_email(email),
                hide_serial(body),
            )
            return None

        return data.get("notifications")

    @staticmethod
    @_catch_all_exceptions
    async def set_notifications(login: AlexaLogin, data) -> dict[str, Any] | None:
        """Update Alexa notification.

        Args:
        login (AlexaLogin): Successfully logged in AlexaLogin
        data : Data to pass to notifications

        Returns json

        """
        response = await AlexaAPI._static_request(
            "put", login, "/api/notifications", data=data
        )
        # _LOGGER.debug("%s: Response: %s", hide_email(login.email),
        #               response.json(content_type=None))
        response_json, *_ = await get_json_value(response, None, dict)
        return response_json

    @staticmethod
    @_catch_all_exceptions
    async def get_dnd_state(login: AlexaLogin) -> dict[str, Any] | None:
        """Get Alexa DND states.

        Args:
        login (AlexaLogin): Successfully logged in AlexaLogin

        Returns json

        """
        response = await AlexaAPI._static_request(
            "get",
            login,
            "/api/dnd/device-status-list",
        )
        response_json, *_ = await get_json_value(response, None, dict)
        return response_json

    @staticmethod
    @_catch_all_exceptions
    async def clear_history(login: AlexaLogin, items: int = 50) -> bool:
        """Clear entries in history."""
        email = login.email
        response = await AlexaAPI._static_request(
            "get", login, "/api/activities", query={"size": items, "offset": -1}
        )

        completed = True
        activities, *_ = await get_json_value(response, "activities", (list, NoneType))
        if not activities:
            _LOGGER.debug("%s: No history to delete.", hide_email(email))
            return True
        _LOGGER.debug(
            "%s:Attempting to delete %s items from history",
            hide_email(email),
            len(activities),
        )
        for activity in activities:
            if not isinstance(activity, dict) or not activity.get("id"):
                continue
            response = await AlexaAPI._static_request(
                "delete",
                login,
                f"/api/activities/{urllib.parse.quote_plus(activity['id'])}",
            )
            if response is None:
                _LOGGER.debug(
                    ("%s:Unable to connect to Alexa to delete %s"),
                    hide_email(email),
                    activity["id"],
                )
            elif response.status == 404:
                _LOGGER.warning(
                    (
                        "%s:Unable to delete %s: %s: \n"
                        "There is no voice recording to delete. "
                        "Please manually delete the entry in the Alexa app."
                    ),
                    hide_email(email),
                    activity["id"],
                    response.reason,
                )
                completed = False
            elif response.status == 200:
                _LOGGER.debug(
                    "%s: Successfully deleted %s",
                    hide_email(email),
                    activity["id"],
                )
        return completed

    @_catch_all_exceptions
    async def set_background(self, url: str) -> bool:
        """Set background for Echo Show.

        Sets the background to Alexa App Photo with the specific https url.

        Args
        url (URL): valid https url for the image

        Returns
        Whether the command was successful.

        """
        data = {
            "deviceSerialNumber": self._device.device_serial_number,
            "deviceType": self._device._device_type,
            "backgroundImageID": "JqIFZhtBTx25wLGTJGdNGQ",
            "backgroundImageType": "PERSONAL_PHOTOS",
            "backgroundImageURL": url,
        }
        _LOGGER.debug(
            "%s: Setting background of %s to: %s",
            hide_email(self._login.email),
            self._device,
            url,
        )
        if url.startswith("http://"):
            _LOGGER.warning("Background URL should be a valid https image")
        response = await self._post_request("/api/background-image", data=data)
        response_json, *_ = await get_json_value(response, None, dict)
        success = bool(response and response.status == 200)
        _LOGGER.debug(
            "%s: Success: %s Response: %s",
            hide_email(self._login.email),
            success,
            response_json,
        )
        return success

    @staticmethod
    @_catch_all_exceptions
    async def force_logout() -> None:
        """Force logout.

        Raises
            AlexapyLoginError: Raise AlexapyLoginError

        """
        raise AlexapyLoginError("Forced Logout")

    @staticmethod
    @_catch_all_exceptions
    async def ping(login: AlexaLogin) -> dict[str, Any] | None:
        """Ping.

        Args:
        login (AlexaLogin): Successfully logged in AlexaLogin

        Returns json

        """
        response = await AlexaAPI._static_request(
            "get",
            login,
            "/api/ping",
        )
        return await response.json(content_type=None) if response else None



"""Python Package for controlling Alexa devices (echo dot, etc) programmatically.

SPDX-License-Identifier: Apache-2.0

Constants.

For more details about this api, please refer to the documentation at
https://gitlab.com/keatontaylor/alexapy
"""

EXCEPTION_TEMPLATE = "An exception of type {0} occurred. Arguments:\n{1!r}"

CALL_VERSION = "2.2.556530.0"
APP_NAME = "Alexa Media Player"

# REST-style Alexa API hosts (for things like /api/notifications)
ALEXA_API_NA = "https://na-api-alexa.amazon.com"
ALEXA_API_EU = "https://eu-api-alexa.amazon.com"
ALEXA_API_FE = "https://fe-api-alexa.amazon.com"

ALEXA_API_BASE = {
    ".com": ALEXA_API_NA,
    ".ca": ALEXA_API_NA,
    ".com.mx": ALEXA_API_NA,
    ".com.br": ALEXA_API_NA,
    ".co.uk": ALEXA_API_EU,
    ".de": ALEXA_API_EU,
    ".fr": ALEXA_API_EU,
    ".es": ALEXA_API_EU,
    ".it": ALEXA_API_EU,
    ".in": ALEXA_API_FE,
    ".co.jp": ALEXA_API_FE,
    ".com.au": ALEXA_API_FE,
    ".co.nz": ALEXA_API_FE,
}

# UA used for REST API calls that should look like a modern mobile client
API_USER_AGENT = (
    "Mozilla/5.0 (Linux; Android 10) "
    "AppleWebKit/537.36 (KHTML, like Gecko) "
    "Chrome/83.0.4103.101 Mobile Safari/537.36"
)

# Default Accept-Language header for REST API calls.
# Amazon documentation and Alexa mobile clients are consistently en-US.
DEFAULT_ACCEPT_LANGUAGE = "en-US"

# Centralize default locale for other operations
DEFAULT_LOCALE = "en-US"

# Retain existing constants (USER_AGENT, etc.) as other code still relies on it.
# These can be removed once all legacy code has been updated to utilize
# the new REST-style constants.
USER_AGENT = f"AmazonWebView/Amazon Alexa/{CALL_VERSION}/iOS/16.6/iPhone"
LOCALE_KEY = {
    ".de": "de_DE",
    ".com.au": "en_AU",
    ".ca": "en_CA",
    ".co.uk": "en_GB",
    ".in": "en_IN",
    ".com": "en_US",
    ".es": "es_ES",
    ".mx": "es_MX",
    ".fr": "fr_FR",
    ".it": "it_IT",
    ".co.jp": "ja_JP",
    ".com.br": "pt_BR",
}
# https://developer.amazon.com/en-US/docs/alexa/alexa-voice-service/api-overview.html#endpoints
HTTP2_NA = "alexa.na.gateway.devices.a2z.com"
HTTP2_EU = "alexa.eu.gateway.devices.a2z.com"
HTTP2_FE = "alexa.fe.gateway.devices.a2z.com"
HTTP2_AUTHORITY = {
    ".com": HTTP2_NA,
    ".ca": HTTP2_NA,
    ".com.mx": HTTP2_NA,
    ".com.br": HTTP2_NA,
    ".co.jp": HTTP2_FE,
    ".com.au": HTTP2_FE,
    ".com.in": HTTP2_FE,
    ".co.nz": HTTP2_FE,
}
HTTP2_DEFAULT = HTTP2_EU

GQL_SMARTHOME_QUERY = """
query CustomerSmartHome {
    endpoints(
      endpointsQueryParams: { paginationParams: { disablePagination: true } }
    ) {
        items {
            legacyAppliance {
                applianceId
                applianceTypes
                friendlyName
                friendlyDescription
                manufacturerName
                connectedVia
                modelName
                entityId
                aliases
                capabilities
                customerDefinedDeviceType
                alexaDeviceIdentifierList
                driverIdentity
            }
        }
    }
}
"""



"""Python Package for controlling Alexa devices (echo dot, etc) programmatically.

SPDX-License-Identifier: Apache-2.0

For more details about this api, please refer to the documentation at
https://gitlab.com/keatontaylor/alexapy
"""

try:
    from importlib_metadata import PackageNotFoundError
    from importlib_metadata import metadata as __load
except ModuleNotFoundError:
    from importlib.metadata import PackageNotFoundError
    from importlib.metadata import metadata as __load

import logging
from pathlib import Path

from .alexaapi import AlexaAPI
from .alexahttp2 import HTTP2EchoClient
from .alexalogin import AlexaLogin
from .alexaproxy import AlexaProxy
from .alexawebsocket import WebsocketEchoClient
from .errors import (
    AlexapyConnectionError,
    AlexapyLoginCloseRequested,
    AlexapyLoginError,
    AlexapyPyotpInvalidKey,
)
from .helpers import hide_email, hide_serial, obfuscate

pkg = Path(__file__).absolute().parent.name
logger = logging.getLogger(pkg)
try:
    metadata = __load(pkg)

    __uri__ = metadata.get("project-url").split("Repository,")[1].strip()
    __title__ = metadata["name"]
    __summary__ = metadata["summary"]
    __license__ = metadata["license"]
    __version__ = metadata["version"]
    __author__ = metadata["author"]
    __maintainer__ = metadata.get("maintainer")
    __contact__ = metadata.get("maintainer")
except PackageNotFoundError:  # pragma: no cover
    logger.error("Could not load package metadata for %s. Is it installed?", pkg)

__all__ = [
    "AlexaAPI",
    "AlexaLogin",
    "AlexaProxy",
    "AlexapyConnectionError",
    "AlexapyLoginCloseRequested",
    "AlexapyLoginError",
    "AlexapyPyotpInvalidKey",
    "HTTP2EchoClient",
    "WebsocketEchoClient",
    "hide_email",
    "hide_serial",
    "obfuscate",
]
